/*
   From "Principles of Parallel Programming", pages 203-204.

   This program uses the following MPI library functions.
     MPI_Init()
     MPI_Comm_size()
     MPI_Comm_rank()
     MPI_Finalize()
     MPI_Send()
     MPI_Recv()
     MPI_Reduce()
*/
#include <stdio.h>
#include <stdlib.h>
#include "mpi.h"

#define RootProcess 0
int length;
int length_per_process;
int myStart;
int myCount=0;
int globalCount;
int tag=1;

int main(int argc, char **argv)
{
   int myID, numProcs, p, i, *myArray;
   FILE* fp;
   MPI_Status status;

   MPI_Init(&argc, &argv);
   MPI_Comm_size(MPI_COMM_WORLD, &numProcs);
   MPI_Comm_rank(MPI_COMM_WORLD, &myID);

   /* Read the data, distribute it among the various processes */
   if(myID == RootProcess)
   {
      if( (fp=fopen(argv[1], "r")) == NULL )
      {
         printf("fopen failed on %s\n", argv[1]);
         return 0;
      }

      fscanf(fp, "%d", &length);  /* read input size */
      length_per_process = length / numProcs;
      /*** allocate space for myArray in the root process ***/
      myArray = (int*)malloc(length_per_process*sizeof(int));

      for(p = 0; p < numProcs-1; p++)  /* read data on behalf of each */
      {                                /* of the other numProcs-1 processes */
         for(i = 0; i < length_per_process; i++)
         {
            fscanf(fp, "%d", myArray+i);
         }
         /*** send length_per_process to the other process ***/
         MPI_Send(&length_per_process, 1, MPI_INT, p+1, tag, MPI_COMM_WORLD);
         /*** send the other process its share of the data ***/
         MPI_Send(myArray, length_per_process, MPI_INT, p+1, tag, MPI_COMM_WORLD);
      }

      for(i = 0; i < length_per_process; i++)  /* now read my data */
      {
         fscanf(fp, "%d", myArray+i);
      }
      fclose(fp);
   }
   else
   {
      /*** receive length_per_process from the root process ***/
      MPI_Recv(&length_per_process, 1, MPI_INT, RootProcess, tag, MPI_COMM_WORLD, &status);
      /*** this process needs to allocate its own space for myArray ***/
      myArray = (int*)malloc(length_per_process*sizeof(int));
      /*** receive my share of the data from the root process ***/
      MPI_Recv(myArray, length_per_process, MPI_INT, RootProcess, tag, MPI_COMM_WORLD, &status);
   }

   /* Do the actual work */
   for(i = 0; i < length_per_process; i++)
   {
      if(myArray[i] == 3)
      {
         myCount++;  /* update local count */
      }
   }

   MPI_Reduce(&myCount, &globalCount, 1, MPI_INT, MPI_SUM, RootProcess, MPI_COMM_WORLD);

   if(myID == RootProcess)
   {
      printf("Number of 3's: %d\n", globalCount);
   }

   MPI_Finalize();
   return 0;
}
